#include "cbvec.h"

///////////////////////////////////////////////////
//                                               //
//                 construction                  //
//                                               //
/////////////////////////////////////////////////// 

vec::vec(int n) : s(n) {
  
  assert(n>0);
  p = new double[s];
  assert(p!=0);                                  // allocation error
  for(int i=0; i<s; i++)
    p[i] = 0;

}

vec::vec(const vec& v) : s(v.s) {
  
  p = new double[s];
  assert(p!=0);
  for(int i=0; i<s; i++)
    p[i] = v.p[i];

}

vec::vec(double a[], int n) : s(n) {

  assert(n>0);
  p = new double[s];
  assert(p!=0);
  for(int i=0; i<s; i++)
    p[i] = a[i];
  
}



///////////////////////////////////////////////////
//                                               //
//               vectoroperations                //
//                                               //
///////////////////////////////////////////////////

// indexing

double vec::getdim(int i) {

   if(i>=s || i<0) {
    cerr << "vec::dim -- out of bounds" << '\n';
    exit(1);
   }
   return p[i];
}  

void vec::setdim(int i, double d) {

   if(i>=s || i<0) {
    cerr << "vec::dim -- out of bounds" << '\n';
    exit(1);
   }
   p[i] = d;

}  

double& vec::operator[](int i) {

   if(i>=s || i<0) {
    cerr << "vec::[] -- out of bounds" << '\n';
    exit(1);
   }
   return p[i];
}  

// assignment

vec& vec::operator=(vec v) {

  if (this !=&v) {
    if(s != v.size()) {
      cerr << "vec::= -- vectors of different size" << '\n';
      exit(1);
    }
    for(int i = 0; i<s; i++)
      p[i] = v.p[i];
  }
  return *this;
  
}

/*
// make a string out of a vector

string vec::tostring() {

  string s1, s2;
  strstream ss;
  for(int i=0; i<s; i++) {
    ss << p[i] << TAB;
    // ss >> s1;
    // s2 += s1;
    // s2 += "\t";
  }

  cerr << ss;

  return s2;

}
*/

// vectoraddition

vec vec::add(vec v) {

  if(s != v.size()) {
    cerr << "vec::add -- vectors of different size" << '\n';
    exit(1);
  }

  vec w(s);

  for(int i=0; i<s; i++)
    w[i]=p[i] + v[i];

  return w;

}

// vectorsubtraction

vec vec::sub(vec v) {

  if(s != v.size()) {
    cerr << "vec::sub -- vectors of different size" << '\n';
    exit(1);
  } 

  vec w(s);

  for(int i=0; i<s; i++)
    w[i]=p[i] - v[i];

  return w;

}

// scalar multiplication

vec vec::mul(double d) {

  vec w(s);

  for(int i=0; i<s; i++)
    w[i] = d * p[i];
  
  return w;

}

// simple euclidean distance

double vec::dis() {

  double d=0;

  for(int i=0; i<s; i++)
    d+=p[i]*p[i];

  return sqrt(d);

}

double vec::dis(vec v) {
  return sub(v).dis();
}


vec operator +(vec v, vec w) {
  return v.add(w);
}

vec operator -(vec v, vec w) {
  return v.sub(w);
}

vec operator *(vec v, double d) {
  return v.mul(d);
}

vec operator *(double d, vec v) {
  return v.mul(d);
}

istream& operator >>(istream& in, vec& v) {
  for(int i=0; i<v.size(); i++)
    in >> v[i];
  return in;
}

ostream& operator <<(ostream& out, vec v) {
  for(int i=0; i<v.size(); i++)
    out << v.getdim(i) << NL;
  return out;
}

///////////////////////////////////////////////////
//                                               //
//     reading/writing vectors from a file       //
//                                               //
///////////////////////////////////////////////////

// read vectors from filestream        
// and create a queue cbvq of these vectors 
// and return it for further processing.
// in case something goes wrong, because the
// file is not of the following form:
// # comment
// DIM DOUBLES + LEARNING RATE
// # comment
// DIM DOUBLES + LEARNING RATE
// ...
// we'll exit -- telling fn.

vector<vec> readvecfile(char* fn) {

  ifstream vf(fn);                               // open vector file
  if(!vf) {
    cerr << "'" << fn << "' "
	 << "readvecfile: file not found" << NL;
    exit(1);
  }
  
  int dim;
  char d[256];                                   // dummy input
  
  vf >> d >> d >> d >> d >> d >> dim;            // read file header

  vector<vec> vv;                                // vector of vecs
  vec vp(dim);                                   // a vector 
  
  string lok;                                    // last okay
  double v;                                      // input variable
  char kflag;                                    // control flag
  
  vf >> kflag, vf.get(d,256);                    // ignore comments
  lok = "top";
  if(kflag==KOMMENTAR) lok=d;                    // last comment that was ok
  
  while(!vf.eof() && !vf.fail()) {               // read rest of file
    
    for(int i=0; i<dim; i++) {                   // go through vector
  
      vf >> v;                                   // read from file
      
      if(vf.eof()  ||                            // if one of those happens
	 vf.fail() ||
	 kflag != KOMMENTAR) {                   // then there is something
	cerr << "readvecfile: '" << fn << "' "   // wrong with the file
	     << "is not okay - ";              
	cerr << "broken at: " 
	     << KOMMENTAR << lok << NL;
	exit(1);
      } else {                                   // otherwise take read double
	vp[i] = v;                               // and put it in vector
      }
    } // end for i
    
    vv.insert(vv.end(), vp);                     // and put new vec in vector
    
    vf >> kflag; vf.get(d,256);                  // ignore comments
    if(kflag==KOMMENTAR) lok=d;                  // last comment that was ok
    
  } // end while
  
  return vv;

}

void writevecfile(vector<vec> vv, char* file) {

  ofstream vf(file);    
  int n = vv.size();
  if(n<1) {
    cerr << "writevecfile: trying to write empty vector<vec>" << NL;
    exit(1);
  } else {
    int dim = vv[0].size();
    
    vf   << KOMMENTAR
	 << " " << n
	 << " vectors of dimension " 
	 << dim << NL << NL;
    
    for(int i=0; i<n; i++) {
      vf   << KOMMENTAR << " vector nr " << i 
	   << " of dimension " << dim << NL;
      vf   << vv[i];
      vf   << NL;
    }
  }

}

//////////////////////////////////////////////////////////////////////////////

istream& operator >>(istream& in, cbvec& v) {
  for(int i=0; i<v.vector.size(); i++)
    in >> v.vector[i];
  in >> v.rate;
  return in;
}

ostream& operator <<(ostream& out, cbvec v) {
  for(int i=0; i<v.vector.size(); i++)
    out << v.vector.getdim(i) << NL;
  out << NL << v.rate;
  return out;
} 

///////////////////////////////////////////////////
//                                               //
//    reading/writing cbvectors from a file      //
//                                               //
///////////////////////////////////////////////////

// read vectors from filestream        
// and create a queue cbvq of these vectors 
// and return it for further processing.
// in case something goes wrong, because the
// file is not of the following form:
// # comment
// DIM DOUBLES + LEARNING RATE
// # comment
// DIM DOUBLES + LEARNING RATE
// ...
// we'll exit -- telling fn.

cbvecqueue readcbfile(char* fn) {

  ifstream cbvf(fn);                             // open codebookvector file
  if(!cbvf) {
    cerr << "'" << fn << "' "
	 << "readcbfile: file not found" << NL;
    exit(1);
  }
  
  int dim;
  char d[256];                                   // dummy input
  
  cbvf >> d >> d >> d >> d >> d >> dim;          // read file header

  cbvecqueue cbq;                                // codebook queue
  cbvec vp(dim);                                 // a cbvec 
  
  string lok;                                    // last okay
  double v;                                      // dummy input
  char kflag;                                    // control flag
  
  cbvf >> kflag, cbvf.get(d,256);                // ignore comments
  lok = "top";
  if(kflag==KOMMENTAR) lok=d;                    // last comment that was ok
  
  while(!cbvf.eof() && !cbvf.fail()) {           // read rest of file
    
    for(int i=0; i<dim; i++) {                   // go through vector
  
      cbvf >> v;                                 // read from file
      
      if(cbvf.eof()  ||                          // if one of those happens
	 cbvf.fail() ||
	 kflag != KOMMENTAR) {                   // then there is something
	cerr << "readcbfile: '" << fn << "' "    // wrong with the file
	     << "is not okay - ";              
	cerr << "broken at: " 
	     << KOMMENTAR << lok << NL;
	exit(1);
      } else {                                   // otherwise take read double
	vp.vector[i] = v;                        // and put it in vector
      }
    } // end for i

    cbvf >> v;                                 // read from file
      
    if(cbvf.eof()  ||                          // if one of those happens
       cbvf.fail() ||
       kflag != KOMMENTAR) {                   // then there is something
      cerr << "readcbfile: '" << fn << "' "    // wrong with the file
	   << "is not okay - ";              
      cerr << "broken at: " 
	   << KOMMENTAR << lok << NL;
      exit(1);
    } else {                                   // otherwise take read double
      vp.rate = v;                             // and set learning rate
    }    
    
    cbq.push(vp);                              // and put new vector in queue
    
    cbvf >> kflag; cbvf.get(d,256);            // ignore comments
    if(kflag==KOMMENTAR) lok=d;                // last comment that was ok
    
  } // end while
  
  return cbq;

}

void writecbfile(cbvecqueue cbq, char* file) {

  ofstream cbvf(file);    
  int n = cbq.size();
  int dim = cbq.front().vector.size();
  
  cbvf << KOMMENTAR
       << " " << n
       << " codebookvectors of dimension " 
       << dim << NL << NL;

  for(int i=0; i<n; i++) {
    cbvf << KOMMENTAR << " cbv nr " << i 
	 << " of dimension " << dim
	 << " -- last component is"
	 << " learning rate" << NL;
    cbvf << cbq.front();
    cbvf << NL << NL;
    cbq.pop();
  }

}

void newcbfile( char* file,                        // new codebookvectorfile
	        int dim, 
	        int n ) {

  ofstream cbvf(file);                          
  
  cbvf << KOMMENTAR
       << " " << n 
       << " codebookvectors of dimension " 
       << dim << NL << NL;

  for(int i=0; i<n; i++) {
    cbvf << KOMMENTAR << " cbv nr " << i 
	 << " of dimension " << dim
	 << " -- last component is"
	 << " learning rate" << NL;
    for(int j=0; j<dim; j++) {
      cbvf << 0 << NL;                           // components 0
    }
    cbvf << NL << 1;                             // learning_rate 1
    cbvf << NL << NL;
  }
}
