#include <stdio.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include "cbvec.h"

#define ERRMSG                                                                \
  	    "\nvq FILENAME [learn/new] [DIM] [N]\n"                           \
	    "   - codebookvectors are loaded from / saved to FILENAME\n"      \
	    "   - 'vq FN new DIM N' creates N new codebookvectors of\n"       \
	    "      dimension DIM and saves them to FN\n"                      \
	    "   - 'vq FN' takes as input doubles from stdin and ouputs\n"     \
	    "     for each DIM-dimensional input vector a N-dimensional\n"    \
	    "     vector of distances to each of the N codebookvectors\n"     \
	    "   - 'vq FN learn' does the same but also modifies the\n"        \
	    "     the codebookvectors\n\n"

#define NEW   (argc == 5)
#define USE   (argc == 2)
#define TRAIN (argc == 3)

typedef queue< double, list<double> > doublequeue;
typedef queue< int, list<int> > intqueue;

intqueue rank(doublequeue dq, 
	      intqueue iq, 
	      double lowerbound, 
	      int r) {
  double min = DBL_MAX;
  unsigned int minpos;

  // das kleinste finden, das noch keinen Wert hat
  for(unsigned int i=0; i<dq.size(); i++) {
      if(iq.front()==0) {
	if(min >= dq.front() && dq.front() >= lowerbound) minpos = i, min=dq.front();
      }
      iq.push(iq.front());
      dq.push(dq.front());
      iq.pop();
      dq.pop();
  }

  // und ihm den neuen rank r geben
  for(unsigned int i=0; i<dq.size(); i++) {
      if(i==minpos) iq.push(r); 
      else iq.push(iq.front());
      dq.push(dq.front());
      iq.pop();
      dq.pop();
  } 
  if(r==(int) dq.size()) return iq;
  else return rank(dq, iq, min, r+1);
}

intqueue rank(doublequeue dq) {
  intqueue iq;
  for(unsigned int i=0; i<dq.size(); i++)
    iq.push(0);
  return rank(dq, iq, -DBL_MAX, 1);
}

///////////////////////////////////////////////////
//                                               //
//           my vector quantisation              //
//                                               //
///////////////////////////////////////////////////

// gets a set of codebookvectors and an input
// vector... returns the modified codebookvectors

// 1. take input datum and measure distance to 
//    each codebook vector in the queue
// 2. sort codebookvectors according to distance
// 3. pull each codebookvector closer to input
// 4. change learning rates
// 5. return new queue of codebookvectors 

cbvecqueue vq(cbvecqueue cbq,                       // codebookvectors 
	      vec iv, int flag) {                   // and an inputvector

  int dim = iv.size();

  doublequeue disq;
  intqueue rankq;

  cbvec cb(dim);
  vec diff(dim);

  // calculate distances to each codebook vector

  double dis = 0;
  double dismin = DBL_MAX;
  for(unsigned int i=0; i<cbq.size(); i++) {
    cb = cbq.front();
    dis = iv.dis(cb.vector);
    if(dis<dismin) dismin = dis;
    disq.push(dis);
    cbq.push(cb);
    cbq.pop();
  }

  if(!flag) {
    for(unsigned int i=0; i<disq.size(); i++) {
      dis = disq.front();
      // if (dis==dismin) cout << i << TAB;
      cout << disq.front() << TAB;
      disq.push(disq.front());
      disq.pop();
    }
    cout << NL;
  }

  // which cb-vector is closed to iv? rank cb-vectors

  rankq = rank(disq);
  int r;

  if(!flag) {
    for(unsigned int i=0; i<rankq.size(); i++) {
      cout << rankq.front() << TAB;
      rankq.push(rankq.front());
      rankq.pop();
    }
    cout << NL;
  }

  // modify cb-vectors

  if(flag) {
    for(unsigned int i=0; i<cbq.size(); i++) {
      
      r    = rankq.front();
      cb   = cbq.front();
      diff = iv - cb.vector;
      cb.vector = cb.vector + diff * pow(cb.rate,(r*r));
      if(r == 1) cb.rate=cb.rate*0.99;
      cbq.push(cb);
      cbq.pop();
      rankq.pop();
      
    }
  }
  
  return cbq;

}
    

///////////////////////////////////////////////////////////////////////////////


int main( int argc, char* argv[] ) {

  if(!( NEW   ||                                 // either we create a new net
        USE   ||                                 // we use a trained net
	TRAIN )) {                               // or we train a net
    cerr << ERRMSG;                              
    return 1;
  }

  char *file;
  int dim, n;
  
  file = argv[1];                                // hopefully good string

  if(NEW) {

    dim = atoi(argv[3]);                         // hopefully proper int
    n = atoi(argv[4]);                           // hopefully proper int

    newcbfile(file, dim, n);

  } else {

    cbvecqueue cbvq;                             // queue of codebookvectors

    cbvq = readcbfile(file);                     // to be loaded from file

    // now the codebookvectors are available and 
    // instances can be read from stdin. this is 
    // what follows. First we have to be told
    // how many dimensions the data have:
    // # many vectors of dimension DIM_I
    // so we can check whether that is okay with
    // the codebookvectors. Then read many ex-
    // amples which look like this:
    // # some comment
    // DIM_I DOUBLES
    // the comment serves as a check symbol to
    // assure that you really serve DIM_I DOUBLES
    // where they were promised.

    int dim = cbvq.front().vector.size();        // dimension of codebook vec.
    int dim_data;                                // dimension of data
    string d;                                    // dummy

    cin >> d >> d >> d >> d >> d >> dim_data;    // read header
    if(dim_data != dim) {
      cerr << "vq: wrong dimensions " 
	   << "codebookvectors have dimension " 
	   << dim << " whereas data has "
	   << dim_data << NL;
      return 1;
    }
   
    char comment[256];
    char cflag;
    char indexfile[256];
    int index = 0;

    vec datum(dim_data);
   
    while(!cin.eof() && !cin.fail()) {           // read as long as you can

      cin >> cflag, cin.get(comment, 256);       // ignore comments and
      cin >> datum;                              // read vector from stream

      if(!cin.eof()) {                           // if we havent run into eof
	if(cflag!=KOMMENTAR) {                   // check if comment is right
	  cerr << "vq: data from stdin is not"
	       << " okay - comment expected" ;
	  return 1;
	}
	if(cin.fail()) {                         // and stream dint fail
	  cerr << "vq: data from stdin is not"
	       << " okay - stdin failed" ;
	   return 1;
	}
      
	if (USE) {
	  cout << KOMMENTAR << comment << NL;
	  cout << index << TAB;
	}
	                                         // everything should be fine: 
	cbvq = vq(cbvq, datum, TRAIN);           // learn !

	sprintf(indexfile, "dummy/cb%d.vq", index);
	writecbfile(cbvq, indexfile);         

	index++;
      }
    } // end while(!cin.eof()...)

    writecbfile(cbvq, file);                     // and write to file again
  
  } // end else if(NEW)

  return 0;

}
